import os
import re

from urllib.parse import (
    ParseResult,
    urlparse, parse_qsl, urlencode, quote, unquote,
 )
from typing import Any, Optional, Tuple, Union, NamedTuple

IP_MIDDLE_OCTET = r"(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5]))"
IP_LAST_OCTET = r"(?:\.(?:0|[1-9]\d?|1\d\d|2[0-4]\d|25[0-5]))"

URL_PATTERN = ( # noqa: W605
    # protocol identifier
    r"(?:(?:https?|ftp)://)"
    # user:pass authentication
    r"(?:[-a-z\u00a1-\uffff0-9._~%!$&'()*+,;=:]+"
    r"(?::[-a-z0-9._~%!$&'()*+,;=:]*)?@)?"
    r"(?:"
    r"(?P<private_ip>"
    # IP address exclusion
    # private & local networks
    r"(?:(?:10|127)" + IP_MIDDLE_OCTET + r"{2}" + IP_LAST_OCTET + r")|"
    r"(?:(?:169\.254|192\.168)" + IP_MIDDLE_OCTET + IP_LAST_OCTET + r")|"
    r"(?:172\.(?:1[6-9]|2\d|3[0-1])" + IP_MIDDLE_OCTET + IP_LAST_OCTET + r"))"
    r"|"
    # private & local hosts
    r"(?P<private_host>"
    r"(?:localhost))"
    r"|"
    # IP address dotted notation octets
    # excludes loopback network 0.0.0.0
    # excludes reserved space >= 224.0.0.0
    # excludes network & broadcast addresses
    # (first & last IP address of each class)
    r"(?P<public_ip>"
    r"(?:[1-9]\d?|1\d\d|2[01]\d|22[0-3])"
    r"" + IP_MIDDLE_OCTET + r"{2}"
    r"" + IP_LAST_OCTET + r")"
    r"|"
    # IPv6 RegEx from https://stackoverflow.com/a/17871737
    r"\[("
    # 1:2:3:4:5:6:7:8
    r"([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|"
    # 1::                              1:2:3:4:5:6:7::
    r"([0-9a-fA-F]{1,4}:){1,7}:|"
    # 1::8             1:2:3:4:5:6::8  1:2:3:4:5:6::8
    r"([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|"
    # 1::7:8           1:2:3:4:5::7:8  1:2:3:4:5::8
    r"([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|"
    # 1::6:7:8         1:2:3:4::6:7:8  1:2:3:4::8
    r"([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|"
    # 1::5:6:7:8       1:2:3::5:6:7:8  1:2:3::8
    r"([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|"
    # 1::4:5:6:7:8     1:2::4:5:6:7:8  1:2::8
    r"([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|"
    # 1::3:4:5:6:7:8   1::3:4:5:6:7:8  1::8
    r"[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|"
    # ::2:3:4:5:6:7:8  ::2:3:4:5:6:7:8 ::8       ::
    r":((:[0-9a-fA-F]{1,4}){1,7}|:)|"
    # fe80::7:8%eth0   fe80::7:8%1
    # (link-local IPv6 addresses with zone index)
    r"fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|"
    r"::(ffff(:0{1,4}){0,1}:){0,1}"
    r"((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}"
    # ::255.255.255.255   ::ffff:255.255.255.255  ::ffff:0:255.255.255.255
    # (IPv4-mapped IPv6 addresses and IPv4-translated addresses)
    r"(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|"
    r"([0-9a-fA-F]{1,4}:){1,4}:"
    r"((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}"
    # 2001:db8:3:4::192.0.2.33  64:ff9b::192.0.2.33
    # (IPv4-Embedded IPv6 Address)
    r"(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])"
    r")\]|"
    # host name
    r"(?:(?:(?:xn--[-]{0,2})|[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]-?)*"
    r"[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]+)"
    # domain name
    r"(?:\.(?:(?:xn--[-]{0,2})|[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]-?)*"
    r"[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]+)*"
    # TLD identifier
    r"(?:\.(?:(?:xn--[-]{0,2}[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]{2,})|"
    r"[a-z\u00a1-\uffff\U00010000-\U0010ffff]{2,}))"
    r")"
    # port number
    r"(?::\d{2,5})?"
    # resource path
    r"(?:/[-a-z\u00a1-\uffff\U00010000-\U0010ffff0-9._~%!$&'()*+,;=:@/]*)?"
    # query string
    r"(?:\?\S*)?"
    # fragment
    r"(?:#\S*)?"
)

pattern = re.compile( r"^" + URL_PATTERN + r"$", re.UNICODE | re.IGNORECASE)

def url_validator(value, public=False):
    """
    Return whether or not givven value is a valid URL.
    If the value is valid URL this function returns ``True``,
    otherwise ``False```.

    This validator is based on `validator of dperini`
    See Also: https://gist.github.com/dperini/729294

    Examples::

        >>> url_validator('http://example.com')
        True

        >>> url_validator('ftp://example.com')
        True

        >>> url_validator('http://example.d')
        False

        >>> url_validator('http://10.0.0.1')
        True

        >>> url_validator('http://10.0.0.1', public=True)
        False

    Parameters
    ----------
    value: str
        URL address string to validate
    public: bool
         If set True to only allow a public IPAddress. (default is False)

    Returns
    -------
    result: bool
        Return ``True`` if the value is valid URL, otherwise ``False``
    """

    check = pattern.match(value)
    if public:
        result = check and not any( check.groupdict().get(key)
                                    for key in ('private_ip', 'private_host'))
    else:
        result = check

    result = True if result else False
    return result


def remove_urls(
        text: str,
        endswith: str ='',
    ) -> str:
    """ Remove any url in the text.
    Parameters
    ----------
        text: str
             The text to remove urls.
        endswith: str
             If set, only remove the URLs that finish with that
             regular expression.
             default is all the URLs are remvoed.
    Returns
    -------
        removed_text: str
            The same text but without urls.
    """
    return replace_urls(text, '', endswith)

def replace_urls(
        text: str,
        replace: str,
        endswith: str = '',
    ) -> str:
    """ Replace all the URLs with path by a text.
    Patameters
    ----------
        text: str
            The text to replace.
        replace: str
            The text to replace with.
        endswith: str
            A regular expression which the URL has to finish with.
            default is replace all the URLs.
    Returns
    -------
        removed_text: str
    Known BUGS
    ----------
        This function is not perfect.
    """
    pattern = re.compile( URL_PATTERN + endswith, re.UNICODE | re.IGNORECASE)
    matches = list(re.finditer(pattern, text))
    matches.reverse()
    for match in matches:
        start, end = match.span()[0], match.span()[1]
        text = text[:start] + replace + text[end:]
    return text


class ResultURLValidator(NamedTuple):
    url: str
    is_valid: bool
    scheme: str
    netloc: str
    username: Optional[str]
    password: Optional[str]
    hostname: str
    port: Optional[str]
    path: str
    params: str
    query: str
    fragment: str
    basename: str


class URL(object):
    __default_safe: str = ':/?&@=#%'
    def __init__(self,
        url: Optional[str]=None,
        do_quote: bool=True,
        safe: Optional[str]=None,
    ):
        f"""
        The class for URL.
        The url is quoted The %-escapes all characters.

        Parameters
        ----------
        url: str
            The any URL
        do_quote: bool
            Replace special characters in string using the %xx escape.
            default is ``True``
        safe: str
            the additional safe chars.
            default safe chars is {self.__default_safe}

        Return
        ------
        url: str
            The url is quoted he %-escapes all characters.
        """
        self.url: str
        self.is_valid: bool
        self.scheme: str
        self.netloc: str
        self.username: Optional[str]
        self.password: Optional[str]
        self.hostname: str
        self.port: Optional[str]
        self.path: str
        self.params: str
        self.query: str
        self.fragment: str
        self.basename: str

        self.safe = safe or self.__default_safe
        if url:
            if do_quote:
                self.url = quote(url, safe=self.safe)
            else:
                self.url = url
        else:
            self.url = None

        self.validate = self.__validator(self.url)
        self.__dict__.update(self.validate._asdict())

    def validator(self, url: str) -> bool:
        """
        Return whether or not given value is a valid URL
        If the value is valid URL this function returns ``True``,
        otherwise ``False``.

        Parameters
        ----------
        url: str
            The input url

        Examples
        --------
        >>> from scrapinghelper import URL
        >>> url = URL()
        >>> url.validator('http://example.com')
        True

        >>> url.validator('ftp://example.com')
        True

        >>> url('http://10.0.0.1')
        True

        """
        return self.__validator(url).is_valid

    def __validator(self,
        url: str
    ) -> ResultURLValidator:

        try:
            v = urlparse(url)
            _is_valid =  url_validator(url)
            _scheme = v.scheme
            _netloc = v.netloc
            _username = v.username
            _password = v.password
            _hostname =  v.hostname
            _port = v.port
            _path = v.path
            _params = v.params
            _query = v.query
            _fragment = v.fragment
            _basename = os.path.basename(unquote(v.path))
        except:
            _is_valid = False
            _scheme = ''
            _netloc = ''
            _username = None
            _password = None
            _hostname =  ''
            _port = None
            _path = ''
            _params = ''
            _query = ''
            _fragment = ''
            _basename = None

        result = ResultURLValidator(
                    url, _is_valid,
                    _scheme, _netloc,
                    _username, _password, _hostname, _port,
                    _path, _params, _query, _fragment, _basename)

        return result

    @property
    def attrs(self):
        return self.validate._asdict()

    def __repr__(self) -> str:
        return str(self.url)

    def unquote(self,
            url: Optional[str]=None,
            **kwargs: Any,
        ) -> str:
        """ Take a unquoted url
        Parameters
        ----------
        url: str
            The input url.
        **kwrags:
            pass to urllib.parse.unquote

        Returns
        -------
        unquoted_url: str
        """

        if not url:
            url = self.url
        return unquote(url, **kwargs)

    def quote(self,
            url: Optional[str]=None,
            **kwargs: Any,
        ) -> str:
        """ Take a quoted url
        Parameters
        ----------
        url: str
            The input url.
        **kwargs:
            pass to urllib.parse.unquote

        Returns
        -------
        quoted_url: str
        """
        if not url:
            url = self.url
        return quote(url, safe=self.safe, **kwargs)

    # alias names
    encode = quote
    decode = unquote

    def set_query_val(self,
            param: str,
            value: Union[str, int, float],
            url: Optional[str]=None,
            update=False,
            use_https=False
        ) -> str:
        """ Takes a url and changes the value of a query string parameter.
        Parameters
        ----------
        param: str
            The name of the query string parameter that needs to be change
        value: Union[str, int, float]
            The new value for the parameter
        create: bool
            if set to True, will create a new query string parameter.
        url: str
            The input url.
        use_https: bool
             If set to true, will upgrade to HTTPS

        Returns
        -------
        Updated URL: str
        """

        if not url:
            url = self.url
            query = self.query

        if not query:
            if "?" in url:
                query = url.split("?")[1]

        qs_val = dict(parse_qsl(query))
        qs_val[f"{param}"] = value
        new_url = url.split("?")[0] + "?" + urlencode(qs_val)

        if use_https:
            self.scheme = "https"
            new_url = new_url.replace("http://", "https://")

        if update:
            self.url = new_url

        return new_url

    def get_query_val(self,
            param: Optional[str]=None,
            url: Optional[str]=None
        ) -> str:
        """Takes a url and extract value of a query string parameter.

        Parameters
        ----------
        url: str
             The input url.
        param: str
             The name of the query string parameter

        Returns
        -------
        val: str
            extract value of a query string parameter
        """

        if not url:
            url = self.url
            query = self.query

        if not query:
            if "?" in url:
                query = url.split("?")[1]

        query_val = dict(parse_qsl(query))
        return query_val.get(param)

    def strip_query(self,
            url: Optional[str]=None
        ) ->str:
        """Takes a url and strips all query string parameters.
        Parameters
        ----------
        url: Any url like https://example.com/sample?src=git

        Returns
        -------
        full url without queries.:
        """

        if not url:
            self.url = f"{self.scheme}://{self.netloc}{self.path}"
            result = self.url
        else:
            v = urlparse(url)
            result = f"{v.scheme}://{v.netloc}{v.path}"
        return result

    def get_root_address(self,
            url: Optional[str]=None
        ) ->str:
        """Takes a url and strips returns the root url
        Parameters
        ----------
        url: str
             Any url like 'https://example.com/sample?src=git'
        full url without parameters: 'https://example.com/'
        """
        if not url:
            result = f"{self.scheme}://{self.netloc}"
        else:
            v = urlparse(url)
            result = f"{v.scheme}://{v.netloc}"
        return result



import os
import re
import random
import itertools
from urllib.parse import urlparse
from pathlib import Path
from typing import Optional, Union, NamedTuple
from enum import Enum
import pandas as pd

IP_MIDDLE_OCTET = r"(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5]))"
IP_LAST_OCTET = r"(?:\.(?:0|[1-9]\d?|1\d\d|2[0-4]\d|25[0-5]))"

PROXY_PATTERN = ( # noqa: W605
    # protocol identifier
    r"(?:(?:(?P<scheme>https?|socks[45]|direct|quick))://)?"
    # user:pass authentication
    r"(?:[-a-z\u00a1-\uffff0-9._~%!$&'()*+,;=:]+"
    r"(?::[-a-z0-9._~%!$&'()*+,;=:]*)?@)?"
    r"(?:"
    r"(?P<private_ip>"
    # IP address exclusion
    # private & local networks
    r"(?:(?:10|127)" + IP_MIDDLE_OCTET + r"{2}" + IP_LAST_OCTET + r")|"
    r"(?:(?:169\.254|192\.168)" + IP_MIDDLE_OCTET + IP_LAST_OCTET + r")|"
    r"(?:172\.(?:1[6-9]|2\d|3[0-1])" + IP_MIDDLE_OCTET + IP_LAST_OCTET + r"))"
    r"|"
    # private & local hosts
    r"(?P<private_host>"
    r"(?:localhost))"
    r"|"
    # IP address dotted notation octets
    r"(?P<public_ip>"
    r"(?:[1-9]\d?|1\d\d|2[01]\d|22[0-3])"
    r"" + IP_MIDDLE_OCTET + r"{2}"
    r"" + IP_LAST_OCTET + r")"
    r"|"
    # host name
    r"(?:(?:(?:xn--[-]{0,2})|[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]-?)*"
    r"[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]+)"
    # domain name
    r"(?:\.(?:(?:xn--[-]{0,2})|[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]-?)*"
    r"[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]+)*"
    # TLD identifier
    r"(?:\.(?:(?:xn--[-]{0,2}[a-z\u00a1-\uffff\U00010000-\U0010ffff0-9]{2,})|"
    r"[a-z\u00a1-\uffff\U00010000-\U0010ffff]{2,}))"
    r")"
    # port number
    r"(?::\d{2,5})?"
)

re_pattern = re.compile( r"^" + PROXY_PATTERN + r"$", re.UNICODE | re.IGNORECASE)

class ProxyRotate( Enum):
    NO_PROXY = 1
    KEEP = 2
    NEXT = 3
    RANDOM = 4

class ProxyParseError(BaseException):
    pass

class ResultProxyValidator(NamedTuple):
    proxy_url: str
    is_valid: bool
    scheme: str
    netloc: str
    username: Optional[str]
    password: Optional[str]
    hostname: str
    port: Optional[str]
    proxy_map: dict

def proxyparse(
        value: str,
        proxy_type='https'
    ) -> ResultProxyValidator:
    """
    Parameters
    ----------
    value: str
        PROXY string to validate
    proxy_type: str
        PROXY type. default is 'https'
    Returns
    -------
    result: ResulttProxyValidator
    """

    if value and not isinstance(value, str):
        value = str(value)

    check = re_pattern.match(value)
    if not check:
        raise ProxyParseError('Invalid proxy string')

    c = any(value.startswith(x) for x in ['http', 'socks', 'direct', 'quic'])
    if not c:
         url_prefix = '{}://'.format(proxy_type) if proxy_type else ''
         proxy_str = '{}{}'.format(url_prefix,value)
    else:
         url_prefix = ''
         proxy_str = value

    proxy_url =  value
    is_valid = True
    v = urlparse(proxy_str)
    scheme = v.scheme
    netloc = v.netloc
    username = v.username
    password = v.password
    hostname = v.hostname
    port = v.port
    proxy_map =  {'http': '{}{}'.format(url_prefix, str(proxy_url)),
                  'https': '{}{}'.format(url_prefix, str(proxy_url)) }

    return ResultProxyValidator( proxy_url, is_valid,
                        scheme, netloc, username, password,
                        hostname, port, proxy_map )

def proxy_validator(value, public=False):
    """
    Return whether or not givven value is a valid PROXY.
    If the value is valid PROXY this function returns ``True``,
    otherwise ``False```.

    This validator is based on `validator of dperini`
    See Also: https://gist.github.com/dperini/729294

    Examples::

        >>> proxy_validator('http://example.com')
        True

        >>> proxy_validator('socks5://127.0.0.1:9050')
        True

        >>> proxy_validator('ftp://example.com')
        False

        >>> proxy_validator('http://example.d')
        False

        >>> proxy_validator('http://10.0.0.1')
        True

        >>> proxy_validator('http://10.0.0.1', public=True)
        False

    Parameters
    ----------
    value: str
        PROXY string to validate
    public: bool
         If set True to only allow a public IPAddress. (default is False)

    Returns
    -------
    result: bool
        Return ``True`` if the value is valid URL, otherwise ``False``
    """

    check = re_pattern.match(value)
    if public:
        result = check and not any( check.groupdict().get(key)
                                    for key in ('private_ip', 'private_host'))
    else:
        result = check

    result = True if result else False
    return result

class PROXY(object):
    def __init__(self,
        proxy_str: Optional[str]=None,
        proxy_type: str='https',
        ) -> ResultProxyValidator:
        self.proxy_url: str
        self.is_valid: bool
        self.scheme: str
        self.netloc: str
        self.username: Optional[str]
        self.password: Optional[str]
        self.hostname: str
        self.port: Optional[str]
        self.proxy_map: dict

        self.validate = self.__validator(proxy_str, proxy_type)
        self.__dict__.update(self.validate._asdict())

    def validator(self, proxy_str: str) -> bool:
        """Validator for PROXY strings.

        Parameters
        ----------
        proxy_str: str
            The input string of proxy.

        Returns
        -------
        Return whether or not given value is a valid PROXY.
        If the proxy_str is valid PROXY this function returns
        ``True``, otherwise ``False``.

        Examples
        --------
        >>> from scrapinghelper import PROXY
        >>> p = PROXY()
        >>> p.validator('127.0.0.1:9050')
        True

        >>> p.validator('ftp://example.com')
        False

        >>> p.validator('http://10.0.0.1')
        True

        >>> p.load_proxies('file:///tmp/myproxies.csv')

        >>> p.load_proxies('https://somewhere.com/share/myproxies.csv')

        >>> p = PROXY('file:///tmp/myproxies.csv')

        >>> p = PROXY('https://somewhere.com/share/myproxies.csv')

        >>> p = PROXY(['127.0.0.1:9050'])

        """
        return self.__validator(proxy_str).is_valid

    def __validator(self,
        proxy_str: str,
        proxy_type: str='https',
    ) -> ResultProxyValidator:

        try:
            v = proxyparse(proxy_str, proxy_type)
            _is_valid =  v.is_valid
            _scheme = v.scheme
            _netloc = v.netloc
            _username = v.username
            _password = v.password
            _hostname =  v.hostname
            _port = v.port
            _proxy_map = v.proxy_map
        except:
            _is_valid = False
            _scheme = ''
            _netloc = ''
            _username = None
            _password = None
            _hostname =  ''
            _port = None
            _proxy_map = {}

        result = ResultProxyValidator(
                    proxy_str, _is_valid,
                    _scheme, _netloc,
                    _username, _password,
                    _hostname, _port, _proxy_map )

        return result

    @property
    def attrs(self):
        return self.validate._asdict()

    def __repr__(self) -> str:
        return str(self.validate.proxy_url)


class ProxyManager(object):
    def __init__(self, proxies: Optional[Union[list,str]]=None):
        """ Proxy Manager
        load proxy data and create proxy pool
        Parameters
        ----------
        proxies: proxies: Optional[Union[list,str]]=None
            list of proxies
            if proxies startswith 'file://', load proxies from file.
            if proxies startswith 'https://', load proxies from URL.
        """
        self._current_proxy: Optional[PROXY] = None
        self._proxies: list = []
        self._proxy_pool: Optional[itertools.cycle]=None
        self._proxy_type: str = 'https'

        proxies = ( proxies
                    or os.environ.get('SCRAPINGHELPER_PROXIES',
                                      default=None) )
        if isinstance(proxies, list):
            self.proxies = proxies
        else:
            if not isinstance(proxies, str):
                proxies = str(proxies)
            if proxies.startswith('file://'):
                filepath = self.normalized_filepath(proxies)
                self.proxies = self.load_proxies(filepath, inplace=False)
            else:
                self.proxies = [proxies]

        self.proxy_pool = itertools.cycle(self.proxies)

    @property
    def proxy_type(self) ->str:
        return self._proxy_type

    @proxy_type.setter
    def proxy_type(self, val):
        if val in ['http', 'https', 'socks4', 'socks5', 'direct', 'quic']:
            self._proxy_type = val
        else:
            raise ValueError('Invalid proxy_type')

    @property
    def proxies(self) ->list:
        return self._proxies

    @proxies.setter
    def proxies(self, val: list):
        if isinstance(val, list) and self._proxies != val:
            self._proxies = val

    @property
    def proxy_pool(self) ->list:
        return self._proxy_pool

    @proxy_pool.setter
    def proxy_pool(self, val: itertools.cycle):
        if isinstance(val, itertools.cycle) and self._proxy_pool != val:
            self._proxy_pool = val

    @property
    def current_proxy(self) ->str:
        if not self._current_proxy:
            self._current_proxy = self.next_proxy(inplace=False)
        return self._current_proxy

    @current_proxy.setter
    def current_proxy(self, val) ->None:
        if self._current_proxy != val:
            self._current_proxy = val

    @classmethod
    def normalized_filepath(cls,
        filepath:str
        ) -> Optional[str]:
        """ normalized filepath.
        Parameters
        ----------
        filepath: str
            if filepath startswith 'file://.',
            expand absolute directory for '.' which is current directory.
        Returns
        -------
        normaized proxies_url:str
        """
        if not filepath:
            return None

        if not isinstance(filepath, str):
            filepath = str(filepath)
        if filepath.startswith('file://.'):
            filepath = filepath.replace('file://.','')
            this_directory = Path(__file__).parent
            filepath = 'file://{}/{}'.format(str(this_directory),filepath)
        return filepath

    def load_proxies(self,
        proxies_url: Optional[str]=None,
        proxy_type: str='https',
        inplace: bool=True,
        ) ->itertools.cycle:
        """Load proxues database from 'proxies_url'.
           and return proxies pool.
        Parameters
        ----------
        proxies_url: str.
            URL of proxies database.
            if proxies_url startswith 'file://.',
            expand absolute directory for '.' which is current directory.
            if proxies_url startswith 'https://',
            read proxies list from URL.

        Returns
        -------
        pool: itertools.cycle
        """

        proxies_url = self.normalized_filepath(proxies_url)
        df = pd.read_csv(proxies_url, names=['proxy'], header=None)
        df['proxy'] = df['proxy'].astype(str)
        proxies =  [ PROXY(x, proxy_type).proxy_url
                          for x in df['proxy'].values.tolist()]
        if inplace:
            self.proxies = proxies
            self.proxy_type = proxy_type
            self.proxy_pool = itertools.cycle(self.proxies)
        else:
            return proxies

    def random_proxy(self, inplace=True) ->PROXY:
        proxy = PROXY(random.choice(self.proxies), self.proxy_type)
        if inplace:
            self.current_proxy = proxy
        return proxy

    def next_proxy(self, inplace=True) ->PROXY:
        proxy =  PROXY(next(self.proxy_pool), self.proxy_type)
        if inplace:
            self.current_proxy = proxy
        return proxy

    def get_proxy(self, rotate: ProxyRotate=ProxyRotate.NEXT) ->PROXY:
        dispatch = { ProxyRotate.NEXT: self.next_proxy,
                     ProxyRotate.RANDOM: self.random_proxy,
                     ProxyRotate.KEEP: lambda : self.current_proxy, }

        if rotate in dispatch:
            return dispatch[rotate]()
        else:
            return None



import sys
from .scraper import (
    Scraper, TAG_LINK, WebScraperException, WebScraperNotFound, MaxRetries,
    HTMLSession, AsyncHTMLSession, HTML, HTMLResponse, Element, PyQuery
)
from .user_agents import UserAgent, user_agent
from .url import URL, remove_urls, replace_urls
from .proxy import ProxyManager, PROXY, ProxyRotate, ProxyParseError
from .logging import logger, LogConfig, LOG_LEVEL
from .versions import __VERSION__

# Sanity checking.
try:
    assert sys.version_info.major == 3
    assert sys.version_info.minor > 7
except AssertionError:
    raise RuntimeError('Requests-HTML requires Python 3.8+!')

__all__ = [
    "Scraper",
    "TAG_LINK",
    "HTMLSession",
    "AsyncHTMLSession",
    "HTML",
    "HTMLResponse",
    "Element",
    "PyQuery",
    "WebScraperException",
    "WebScraperNotFound",
    "MaxRetries",
    "user_agent",
    "UserAgent",
    "URL",
    "remove_urls",
    "replace_urls",
    "ProxyManager",
    "ProxyRotate",
    "ProxyParseError",
    "PROXY",
    "logger",
    "LogConfig",
    "LOG_LEVEL",
    "__VERSION__",
]
